#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glaux.h>
#include <GL/glut.h>

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

#define DEBUG 0
#define QT_TEXTURAS 3

/* VARIAVEIS GLOBAIS */

typedef struct {
  GLboolean   doubleBuffer;
  GLint       delay;
  GLboolean	usaLuz;
}Estado;

typedef struct {
  GLfloat     ang_h;
  GLfloat     ang_v;
  GLfloat     dist;
}Camara;

typedef struct {
	GLuint textura[QT_TEXTURAS];
	GLboolean esfera;
	GLboolean usaTextura;
}Modelo;

Estado estado;
Camara camara;
Modelo modelo;


typedef struct {
  int         sizeX, sizeY, bpp;
  char        *data;
}JPGImage;

typedef struct {
  int         sizeX, sizeY;
  char        *data;
} PPMImage;

extern "C" int read_JPEG_file(char *, char **, int *, int *, int *);
extern "C" PPMImage *LoadPPM(char *);

/* Inicializao do ambiente OPENGL */
void GerarTexturas(void)
{
   AUX_RGBImageRec *imagemBMP;
   JPGImage imagemJPG;
   PPMImage *imagemPPM;

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   glGenTextures(QT_TEXTURAS, modelo.textura);

   imagemBMP = auxDIBImageLoad("textura.bmp");
   glBindTexture(GL_TEXTURE_2D, modelo.textura[0]);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
   gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, imagemBMP->sizeX, imagemBMP->sizeY, GL_RGB, GL_UNSIGNED_BYTE, imagemBMP->data);

   (void) read_JPEG_file("textura.jpg", &imagemJPG.data, &imagemJPG.sizeX, &imagemJPG.sizeY, &imagemJPG.bpp);
   glBindTexture(GL_TEXTURE_2D, modelo.textura[1]);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
   gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, imagemJPG.sizeX, imagemJPG.sizeY, GL_RGB, GL_UNSIGNED_BYTE, imagemJPG.data);

   imagemPPM = LoadPPM("textura.ppm");
   glBindTexture(GL_TEXTURE_2D, modelo.textura[2]);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
   gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, imagemPPM->sizeX, imagemPPM->sizeY, GL_RGB, GL_UNSIGNED_BYTE, imagemPPM->data);

   // psousa -->
   AUX_RGBImageRec *imagemBMP256;
   imagemBMP256 = auxDIBImageLoad("tex256.bmp");
   glBindTexture(GL_TEXTURE_2D, modelo.textura[3]);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
   gluBuild2DMipmaps(GL_TEXTURE_2D, GL_R3_G3_B2, imagemBMP256->sizeX, imagemBMP256->sizeY, GL_RGB, GL_UNSIGNED_BYTE, imagemBMP256->data);
   // <-- psousa

   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void Init(void)
{
  glClearColor(0.0, 0.0, 0.0, 0.0);

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_POINT_SMOOTH);
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_POLYGON_SMOOTH);
  camara.ang_h = 45.0;
  camara.ang_v = 45.0;
  camara.dist = 3.0;
  GerarTexturas();


}

/**************************************
***  callbacks de janela/desenho    ***
**************************************/

// CALLBACK PARA REDIMENSIONAR JANELA

void Reshape(int width, int height)
{
  GLint size;

  if (width < height)
    size = width;
  else
    size = height;

  // glViewport(botom, left, width, height)
  // define parte da janela a ser utilizada pelo OpenGL

  glViewport(0, 0, (GLint) size, (GLint) size);
  

  // Matriz Projeccao
  // Matriz onde se define como o mundo e apresentado na janela
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  // gluOrtho2D(left,right,bottom,top); 
  // projeccao em perspectiva
  gluPerspective(45.0, 1.0, 0.1, 10.0);

  // Matriz Modelview
  // Matriz onde so realizadas as tranformacoes dos modelos desenhados
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}


// ... definicao das rotinas auxiliares de desenho ...

void desenhaPoligono(GLfloat a[], GLfloat b[], GLfloat c[], GLfloat  d[], GLfloat cor[])
{
  glBegin(GL_POLYGON);
    glColor3fv(cor);
	glTexCoord2f(0.0, 0.0);
    glVertex3fv(a);
	glTexCoord2f(1.0, 0.0);
    glVertex3fv(b);
	glTexCoord2f(1.0, 1.0);
    glVertex3fv(c);
	glTexCoord2f(0.0, 1.0);
    glVertex3fv(d);
  glEnd();
}

void desenhaCubo(void)
{
  GLfloat vertices[][3] = {
				{-0.5,-0.5,-0.5}, 
                {0.5,-0.5,-0.5}, 
                {0.5,0.5,-0.5}, 
                {-0.5,0.5,-0.5}, 
                {-0.5,-0.5,0.5},  
                {0.5,-0.5,0.5}, 
                {0.5,0.5,0.5}, 
                {-0.5,0.5,0.5}
  };

  GLfloat cores[][3] = {
						{0.0,1.0,1.0},
                        {1.0,0.0,0.0},
                        {1.0,1.0,0.0}, 
                        {0.0,1.0,0.0}, 
                        {1.0,0.0,1.0}, 
                        {0.0,0.0,1.0}, 
                        {1.0,1.0,1.0}
  };

  glBindTexture(GL_TEXTURE_2D, modelo.textura[0]);
  desenhaPoligono(vertices[1],vertices[0],vertices[3],vertices[2],cores[0]);
  desenhaPoligono(vertices[2],vertices[3],vertices[7],vertices[6],cores[1]);
  glBindTexture(GL_TEXTURE_2D, modelo.textura[1]);
  desenhaPoligono(vertices[3],vertices[0],vertices[4],vertices[7],cores[2]);
  desenhaPoligono(vertices[6],vertices[5],vertices[1],vertices[2],cores[3]);
  glBindTexture(GL_TEXTURE_2D, modelo.textura[2]);
  desenhaPoligono(vertices[4],vertices[5],vertices[6],vertices[7],cores[4]);
  desenhaPoligono(vertices[5],vertices[4],vertices[0],vertices[1],cores[5]);
}

// Callback de desenho

void Draw(void)
{
  // psousa -->
    if (modelo.usaTextura)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);
	// <-- psousa

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
// ... transformacoes de visualizacao ...
  glPushMatrix();
  glTranslatef(0.0, 0.0, -camara.dist);
  glRotatef(camara.ang_v, 1.0, 0.0, 0.0);
  glRotatef(camara.ang_h, 0.0, 1.0, 0.0);
// ... chamada das rotinas auxiliares de desenho ...

  // psousa -->
	if (estado.usaLuz)
	{
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);

	float pos[] = {2, 2, 2, 0};
	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	}
	else
	{
		glDisable(GL_LIGHTING);
	}
	if (!modelo.esfera)
	{
  desenhaCubo();
	}
	else
	{
  glBindTexture(GL_TEXTURE_2D, modelo.textura[0]);

  float amb[] = {1, 1, 1, 1};
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb);
  GLUquadricObj* l_poQuadric = gluNewQuadric();
gluQuadricDrawStyle(l_poQuadric, GLU_FILL);
gluQuadricNormals(l_poQuadric, GLU_SMOOTH);
if (modelo.usaTextura)
	gluQuadricTexture(l_poQuadric, GL_TRUE); 
gluSphere(l_poQuadric, 1, 30, 30);
gluDeleteQuadric(l_poQuadric);
	}
	// <-- psousa

  glPopMatrix();
  glFlush();
  if (estado.doubleBuffer)
    glutSwapBuffers();
}

/*******************************
***   callbacks timer/idle   ***
*******************************/

// Callback Idle

void Idle(void)
{

  // accoes a tomar quando o glut est idle 

  // redesenhar o ecra 
  // glutPostRedisplay();
}

// Callback de temporizador

void Timer(int value)
{
  glutTimerFunc(estado.delay, Timer, 0);
  // ... accoes do temporizador ... 


  // redesenhar o ecra 
  glutPostRedisplay();
}


void imprime_ajuda(void)
{
  printf("\n\nImportacao de um modelo 3DS\n");
  printf("h,H - Ajuda \n");
  printf("Setas  - Rotacao da camara\n");
  printf("PageUp,PageDown - Distancia da camara ao cubo\n");
  printf("ESC - Sair\n");
}

/*******************************
***  callbacks de teclado    ***
*******************************/

// Callback para interaccao via teclado (carregar na tecla)

void Key(unsigned char key, int x, int y)
{
  switch (key) {
    case 27:
      exit(1);
    // ... accoes sobre outras teclas ... 

	  // psousa -->
	case ' ':
		modelo.esfera = !modelo.esfera;
		glutPostRedisplay();
		break;
	case 'l':
	case 'L':
		estado.usaLuz = !estado.usaLuz;
		printf("Iluminacao: %s\n", (estado.usaLuz ? "ON" : "off"));
		glutPostRedisplay();
		break;
	case 't':
	case 'T':
		modelo.usaTextura = !modelo.usaTextura;
		printf("Textura: %s\n", (modelo.usaTextura ? "ON" : "off"));
		glutPostRedisplay();
		break;

	  // <-- psousa

    case 'h' :
    case 'H' :
                imprime_ajuda();
            break;
  }

  if(DEBUG)
    printf("Carregou na tecla %c\n",key);

}

// Callback para interaccao via teclado (largar a tecla)

/*void KeyUp(unsigned char key, int x, int y)
{

  if(DEBUG)
    printf("Largou a tecla %c\n",key);
}*/

// Callback para interaccao via teclas especiais  (carregar na tecla)

void SpecialKey(int key, int x, int y)
{
  // ... accoes sobre outras teclas especiais ... 
  //    GLUT_KEY_F1 ... GLUT_KEY_F12
  //    GLUT_KEY_UP
  //    GLUT_KEY_DOWN
  //    GLUT_KEY_LEFT
  //    GLUT_KEY_RIGHT
  //    GLUT_KEY_PAGE_UP
  //    GLUT_KEY_PAGE_DOWN
  //    GLUT_KEY_HOME
  //    GLUT_KEY_END
  //    GLUT_KEY_INSERT 

  switch (key) {
	case GLUT_KEY_LEFT:
		camara.ang_h -= 5;
		break;
	case GLUT_KEY_RIGHT:
		camara.ang_h += 5;
		break;
	case GLUT_KEY_DOWN:
		camara.ang_v -= 5;
		break;
	case GLUT_KEY_UP:
		camara.ang_v += 5;
		break;
	case GLUT_KEY_PAGE_UP:
		camara.dist /= 1.1;
		break;
	case GLUT_KEY_PAGE_DOWN:
		camara.dist *= 1.1;
		break;
  }
  if(DEBUG)
    printf("angH = %f graus, angV = %f graus, Dist = %f\n", camara.ang_h, camara.ang_v, camara.dist);
  // redesenhar o ecra 
  glutPostRedisplay();

  if(DEBUG)
    printf("Carregou na tecla especial %d\n",key);
}

// Callback para interaccao via teclas especiais (largar na tecla)

/*void SpecialKeyUp(int key, int x, int y)
{

  if(DEBUG)
    printf("Largou a tecla especial %d\n",key);

}*/


void main(int argc, char **argv)
{
  estado.doubleBuffer=GL_TRUE;

  glutInit(&argc, argv);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(300, 300);
  glutInitDisplayMode((estado.doubleBuffer) ? GLUT_DOUBLE : GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
  if (glutCreateWindow("Exemplo") == GL_FALSE)
    exit(1);

  Init();

  imprime_ajuda();

// Registar callbacks do GLUT
  
  // callbacks de janelas/desenho
  glutReshapeFunc(Reshape);
  glutDisplayFunc(Draw);

  // Callbacks de teclado
  glutKeyboardFunc(Key);
  //glutKeyboardUpFunc(KeyUp);
  glutSpecialFunc(SpecialKey);
  //glutSpecialFunc(SpecialKeyUp);

  // callbacks timer/idle
  //glutTimerFunc(estado.delay, Timer, 0);
  //glutIdleFunc(Idle);

  
  // COMECAR...
  glutMainLoop();
}
